/*
 * Tabs are 4 symbols wide due to stupid Windows function names...
 * ---------------------------------------------------------------
 *
 * Copyright (c) 2010 KSI Labs, LLC. All rights reserved.
 * Written by Sergey Kubushyn <ksi@koi8.net>, <ksi@ksilabs.com>.
 *
 * File: mach3.c
 *
 * This is Mach3-specific routines for  WDM driver for KSI Labs, LLC
 * CENTIPEDE PCI Board.
 *
 * Please read documentation or go to http://www.ksilabs.com for further info.
 */

/*
 * 						LICENSE and Terms of Use.
 *
 * First of all, this is NOT GPL or Free Software. It is open so you have
 * all the source code for the Engine. There is also VHDL source code for
 * CENTIPEDE board CPLD that is available from KSI Labs, LLC website,
 * http://www.ksilabs.com.
 *
 * As for this particular code, Windows Mach3 Engine, you are free to use it,
 * make any modifications to customize it for your own needs, fix inevitable
 * bugs, port it to other CNC software and do anything else as long as your
 * derivative work that is based on this code or uses parts or pieces thereof
 * is made to be used with any KSI Labs, LLC products.
 *
 * We do NOT grant you a permission to use any parts of this software in any
 * products that are designed for products not manufactured or distributed
 * by KSI Labs, LLC. We might give such a permission on case-by-case basis
 * but it is not legal to use this software or bits and pieces thereof with
 * products not coming from KSI Labs, LLC without our explicit written
 * permission.
 *
 * The original copyright notices must be included in all derivative works.
 * You do NOT have to ask our permission to make any modifications to this
 * code and you are not under obligation to give us any of your work back
 * as long as you are using it for your own needs.
 *
 * However we do NOT give our permission to distribute your derivative work
 * unless all original copyrights are retained, all the conditions of this
 * license are met (use only with KSI Labs, LLC products) and full source
 * code for such derivative work is available for free to everybody without
 * any further conditions including but not limited to signing any kinds of
 * NDA. We explicitely deny permission to distribute any derivative work
 * based on this code with closed source binary-only fragments or include
 * bits or pieces of this code in such products.
 *
 * CNC software like Mach3 and likes is excluded from these requirements as
 * long as it does not include parts or pieces of this software into the
 * close source proprietary application proper. It can use shared memory
 * buffer, IOCTL calls, etc. to communicate with this Engine or the one based
 * on it and keep all the software making those calls or working on that
 * buffer close source and proprietary as long as those calls/structures are
 * documented and that documentation is freely available to everybody without
 * any further conditions including but no limited to signing any kind of NDA.
 */

#include "precomp.h"

#define INSIG_SIZE	16
#define OUTSIG_SIZE	5
#define PFUNC_SIZE	4


// ***********************************************************************
// * This maps Input and Output signals to Parallel Ports/CENTIPEDE pins *
// * by populating GetSigMap[] and PutSigMap[] Call Tables with pointers *
// * to the proper functions that come from sigfuncs.c. Also prepares    *
// * Parallel Ports for operation in Input Mode if they configured so.   *
// ***********************************************************************
static VOID MachMapIOSignals(PFDO_DATA FdoData)
{
	UINT32					tmp;
	INT						i;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	pGetSigFunc				*GetSigMap = FdoData->GetSigMap;
	pPutSigFunc				*PutSigMap = FdoData->PutSigMap;
	pGetSigFunc				*GetMPGSigMap = FdoData->GetMPGSigMap;

	FdoData->Port1isInput = FdoData->Port2isInput = FALSE;
	FdoData->Port1On = FdoData->Port2On = FALSE;
	FdoData->PortsMapOK = TRUE;

	// ###########################
	// Set input signals mapping #
	// ###########################

	//
	// Fixed Limit/Home switches, per axis. We override Mach3 config
	// for those. InPort/InPin/Active overriden, Negated taken from Mach3.
	// We do also override per-axis Step/Dir Pin/Port for all axes here.
	// They should be setup in Mach3 config but anyways...
	//
	for (i = 0; i < 6; i++) {

		if (mach->Axis[i].Enable) {
			//
			// Home/Limit switches. They are NOT just signals to Mach3,
			// they are handled by CENTIPEDE hardware so they MUST be
			// setup properly. Those can NOT be reassigned arbitrarily.
			//
			mach->InSigs[i*3].Active = TRUE;
			mach->InSigs[i*3].InPort = CPD_PORT;
			mach->InSigs[i*3].InPin = i * 3 + 1;
			mach->InSigs[i*3+1].Active = TRUE;
			mach->InSigs[i*3+1].InPort = CPD_PORT;
			mach->InSigs[i*3+1].InPin = i * 3 + 2;
//			mach->InSigs[i*3+2].Active = TRUE;
			mach->InSigs[i*3+2].InPort = CPD_PORT;
			mach->InSigs[i*3+2].InPin = i * 3 + 3;
			//
			// Step/Dir outputs. They are fixed per axis so they MUST be
			// setup properly. Those can NOT be reassigned arbitrarily.
			// Those are NOT mapped to any OutSigs[]; they are special
			// and they are not processed by MachPutOutputs().
			//
			mach->Axis[i].StepPort = mach->Axis[i].DirPort = CPD_PORT;
			mach->Axis[i].DirPin = i * 2 + 1;
			mach->Axis[i].StepPin = i * 2 + 2;
			//
			// Map input signals functions to proper pins.
			//
			GetSigMap[i*3] = (mach->InSigs[i*3].Negated) ?
							nInCPDMap[i*3] : InCPDMap[i*3];
			GetSigMap[i*3+1] = (mach->InSigs[i*3+1].Negated) ?
							nInCPDMap[i*3+1] : InCPDMap[i*3+1];
			GetSigMap[i*3+2] = (mach->InSigs[i*3+2].Negated) ?
							nInCPDMap[i*3+2] : InCPDMap[i*3+2];
		} else {
			//
			// Axis is disabled so disable all of its input signals.
			//
			mach->InSigs[i*3].Active = FALSE;
			mach->InSigs[i*3].InPort = 0;
			mach->InSigs[i*3].InPin = 0;
			mach->InSigs[i*3+1].Active = FALSE;
			mach->InSigs[i*3+1].InPort = 0;
			mach->InSigs[i*3+1].InPin = 0;
			mach->InSigs[i*3+2].Active = FALSE;
			mach->InSigs[i*3+2].InPort = 0;
			mach->InSigs[i*3+2].InPin = 0;
			//
			// Map corresponding signal functions to NULL so they will
			// be skipped in MachGetInputs (that will make them unActivated.)
			//
			GetSigMap[i*3] = NULL;
			GetSigMap[i*3+1] = NULL;
			GetSigMap[i*3+2] = NULL;
			//
			// Step/Dir outputs. They are fixed per axis so they MUST be
			// setup properly. Those can NOT be reassigned arbitrarily.
			// They actually don't matter when Axis is disabled but set
			// them to zero just for consistency.
			//
			mach->Axis[i].StepPort = mach->Axis[i].DirPort = 0;
			mach->Axis[i].DirPin = mach->Axis[i].StepPin = 0;
		}
	}

	//
	// Step/Dir Spindle. We override Mach3 config for it. There are no
	// Inputs for this--no Limit/Home switches on the Spindle--only Outputs.
	// We do override Step/Dir Pin/Port for the Spindle here. They should
	// be setup in Mach3 config but anyways...
	//
	if (mach->Axis[SPINDLE].Enable) {
		//
		// Step/Dir outputs. They are fixed per axis so they MUST be
		// setup properly. Those can NOT be reassigned arbitrarily.
		// Those are NOT mapped to any OutSigs[]; they are special
		// and they are not processed by MachPutOutputs().
		//
		mach->Axis[SPINDLE].StepPort = mach->Axis[SPINDLE].DirPort = CPD_PORT;
		mach->Axis[SPINDLE].DirPin = 13;
		mach->Axis[SPINDLE].StepPin = 14;
	} else {
		//
		// Step/Dir outputs. They are fixed so they MUST be setup
		// properly. Those can NOT be reassigned arbitrarily.
		// They actually don't matter when Step/Dir Spindle is
		// disabled but set them to zero just for consistency.
		//
		mach->Axis[SPINDLE].StepPort = mach->Axis[SPINDLE].DirPort = 0;
		mach->Axis[SPINDLE].DirPin = mach->Axis[SPINDLE].StepPin = 0;
	}

	//
	// Fixed Index and Timing inputs. Pin/Port are fixed, everything else
	// comes from Mach3 config.
	//
	mach->InSigs[TIMING].InPort = CPD_PORT;
	mach->InSigs[TIMING].InPin = 23;
	mach->InSigs[INDEX].InPort = CPD_PORT;
	mach->InSigs[INDEX].InPin = 24;
	if (mach->InSigs[INDEX].Active) {
		GetSigMap[INDEX] = (mach->InSigs[INDEX].Negated) ?
					nInCPDMap[23] : InCPDMap[23];
	} else {
		GetSigMap[INDEX] = NULL;
	}
	if (mach->InSigs[TIMING].Active) {
		GetSigMap[TIMING] = (mach->InSigs[TIMING].Negated) ?
					nInCPDMap[22] : InCPDMap[22];
	} else {
		GetSigMap[TIMING] = NULL;
	}

	//
	// Probe and EStop are also fixed. Mach3 config also overriden.
	// For EStop we force Negated so EStop switch is activated when open.
	//
	mach->InSigs[DIGITIZE].Active = TRUE;
	mach->InSigs[DIGITIZE].InPort = CPD_PORT;
	mach->InSigs[DIGITIZE].InPin = 19;
	mach->InSigs[EMERGENCY].Active = TRUE;
	mach->InSigs[EMERGENCY].Negated = TRUE;
	mach->InSigs[EMERGENCY].InPort = CPD_PORT;
	mach->InSigs[EMERGENCY].InPin = 20;
	GetSigMap[DIGITIZE] = (mach->InSigs[DIGITIZE].Negated) ?
					nInCPDMap[18] : InCPDMap[18];
	GetSigMap[EMERGENCY] = nInCPDMap[19];	// EStop MUST be NC switch !!!

	//
	// Fixed Encoders inputs. They are not signals in Mach3 sense and they
	// are not used for anything by it but we override Mach3 config
	// for those. They are set in Encoders[7] structures as APin/APort and
	// BPin/BPort. We do only override first 4 encoders that are supposed
	// to work with relatively fast linear scales (or something similar)
	// so they have hardware assistance in CENTIPEDE. These mappings only
	// serve informational purpose; they are displayed in Config Dialog.
	//
	// Remaining 3 Encoders are for relatively slow MPGs so they get
	// all-software processing and can be mapped to any input pins including
	// Parallel Ports. They DO use our regular I/O processing routines for
	// getting signals from MPGs and the actual processing is done in
	// MachUpdateMPGs() routine.
	//
	for (i = 0; i < 4; i++) {
		if (mach->Encoders[i].Active) {
			mach->Encoders[i].APort = mach->Encoders[i].BPort = CPD_PORT;
			mach->Encoders[i].APin = 25 + i*2;
			mach->Encoders[i].BPin = 25 + i*2 + 1;
		} else {
			mach->Encoders[i].APort = mach->Encoders[i].BPort = 0;
			mach->Encoders[i].APin = mach->Encoders[i].BPin = 0;
		}
	}

	//
	// Rest of inputs. Those are taken from Mach3.
	//
	for (i = 18; i < nSigs; i++) {	// Limit/Home Switches skipped
		if (i == DIGITIZE) continue;	// Probe Switch skipped (fixed)
		if (i == TIMING) continue;		// Spindle Timing skipped (fixed)
		if (i == INDEX) continue;		// Spindle Index skipped (fixed)
		if (i == EMERGENCY) continue;	// EStop Switch -- also fixed

		if (!mach->InSigs[i].Active) {
			GetSigMap[i] = NULL;
			continue;
		}

		MapOneInput(FdoData, &GetSigMap[i], mach->InSigs[i].InPort,
						mach->InSigs[i].InPin, mach->InSigs[i].Negated);
	}

	//
	// One more time for MPG inputs. No Negate, they are taken as-is.
	//
	for (i = 0; i < 3; i++) {			// We have 3 MPGs in Mach3

		if (!mach->Encoders[i+4].Active) {
			GetMPGSigMap[i*2] = NULL;		// 2 signals per MPG,
			GetMPGSigMap[i*2+1] = NULL;		// A and B phases
			continue;
		}

		MapOneInput(FdoData, &GetMPGSigMap[i*2], mach->Encoders[i+4].APort,
											mach->Encoders[i+4].APin, FALSE);

		MapOneInput(FdoData, &GetMPGSigMap[i*2+1], mach->Encoders[i+4].BPort,
											mach->Encoders[i+4].BPin, FALSE);
	}

	// ############################
	// Set output signals mapping #
	// ############################

	for (i = 0; i < nSigsOut; i++) {

		MapOneOutput(FdoData, &PutSigMap[i], mach->OutSigs[i].OutPort,
											mach->OutSigs[i].OutPin);
	}
}


// **************************************************************
// * This is MACH1_STARTTIMER IOCTL handler. It is invoked with *
// * Mach3 structure already mapped to kernel address space so  *
// * we do not have to check if it is mapped. Sets up all data  *
// * structures required to run our Engine based on Mach3 and   *
// * its own configuration, maps I/O signals to pins, acquires  *
// * system Parallel Ports if it is configured to use them,     *
// * configures them to Input mode if requested and starts the  *
// * board in EStop mode (only periodic timer is running.)      *
// **************************************************************
BOOLEAN MachDoIoctlStartTimer(PFDO_DATA FdoData)
{
	UINT32						tmp, dummy;
	INT							i;
	volatile PMACH_STRUCT		mach = FdoData->MachSharedBlock;
	volatile PAXIS_DATA			AxData = FdoData->AxisData;
	PPARALLEL_PORT_INFORMATION	pports = FdoData->PPortInfo;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "-->MachDoIoctlStartTimer()\n");
#endif

	//
	// Start in EStop, 25KHz. That won't let any steps out,
	// FIFO is not popped, just periodic timer will be running.
	// We'll have IRQ_FAULT set until the EStop is gotten rid of,
	// switches IRQs if enabled (they are not at this stage,)
	// IRQ_FOVF (it should never happen--VHDL error if it is
	// raised,) and periodic timer IRQs every 200uS (40 x 5.)
	//
	// All IRQs are masked so far (by board reset) so no ISR
	// invoked.
	//

	// ##########################################
	// mach_cfg* registers, AxisData[], Spindle #
	// ##########################################
	dummy = tmp = 0;
	for (i = 0; i < 6; i++) {
		AxData[i].Active = FALSE;
		AxData[i].Slave = -1;
		AxData[i].Master = -1;
		AxData[i].Type = Single;
		AxData[i].Hit = FALSE;
		AxData[i].DeRefAdjust = FALSE;
		AxData[i].DoAdjust = FALSE;
		AxData[i].QuantCntr = 0;
	}
	for (i = 0; i < 6; i++) {
		// AxisData[]
		if (mach->Axis[i].Slave && (mach->Axis[i].SlaveAxis < 6) &&
											(mach->Axis[i].SlaveAxis >= 0)) {
				AxData[mach->Axis[i].SlaveAxis].Slave = i;
				AxData[mach->Axis[i].SlaveAxis].Type = Master;
				AxData[i].Master = mach->Axis[i].SlaveAxis;
				AxData[i].Type = Slave;
		}
		// Axes enables
		tmp |= (mach->Axis[i].Enable) ? 1 << (i*5) : 0;
		// Axes dir reversed (2 variables, Axis[i].DirNegate and Directions[i])
		tmp |= (mach->Axis[i].DirNegate != mach->Directions[i]) ? 1 << (i*5+1) : 0;
		// Axes Step Signal Inverted
		dummy |= (mach->Axis[i].StepNegate) ? 1 << (i) : 0;
		// Axes home switches NO (Normally Open)
		tmp |= (mach->InSigs[i*3+2].Negated) ? 0 : 1 << (i*5+3);
		// Axes limit switches NO (Normally Open)
		tmp |= (mach->InSigs[i*3].Negated || mach->InSigs[i*3+1].Negated) ?
															0 : 1 << (i*5+2);
	}
	// Probe switch NO (Normally Open)
	tmp |= (mach->InSigs[DIGITIZE].Negated) ? 0 : PROBE_SW_NO;

	//
	// Setup Spindle Index if enabled
	//
	if (mach->InSigs[INDEX].Active) {
		dummy |= SP_IDX_EN;
		if (mach->InSigs[INDEX].Negated) dummy |= SP_IDX_REV;
		dummy &= ~IDX_DEBOUNCE_MASK;
		dummy |= IDX_DEBOUNCE_BITS(mach->SpinDebounce);
	}

	//
	// Setup Spindle Timing if enabled
	//
	if (mach->InSigs[TIMING].Active) {
		dummy |= SP_TIMING_EN;
		if (mach->InSigs[TIMING].Negated) dummy |= SP_TIMING_REV;
		dummy &= ~TIM_DEBOUNCE_MASK;
		dummy |= TIM_DEBOUNCE_BITS(mach->Debounce);
	}

	//
	// Step/Dir spindle...
	//
	if (mach->Axis[SPINDLE].Enable) {
		dummy |= SP_DIRSTEP;
		if (mach->Axis[SPINDLE].StepNegate) dummy |= SP_STEP_REV;
		if (mach->Axis[SPINDLE].DirNegate) dummy |= SP_DIR_REV;
		mach->Axis[SPINDLE].MaxVelocity = mach->Axis[SPINDLE].CurVelocity = 0;
	}

	// Now write it out (Home and Probe switches disabled for now)
	FdoData->CPDRegsBase->mach_cfg = tmp;
	FdoData->CPDRegsBase->mach_cfg2 = dummy;
	tmp = FdoData->CPDRegsBase->mach_cfg;

	//
	// Build I/O signals function pointers maps according to Mach3 config
	//
	MachMapIOSignals(FdoData);

	//
	// Set Parallel Ports Direction Bits if they are in Input mode
	//
	if (FdoData->PortsMapOK) {
		if (FdoData->Port1isOK && FdoData->Port1On && FdoData->Port1isInput) {
			tmp = READ_PORT_UCHAR(pports[0].Controller + 2);
			tmp |= 0x20;
			WRITE_PORT_UCHAR(pports[0].Controller + 2, (UCHAR)tmp);
		}
		if (FdoData->Port2isOK && FdoData->Port2On && FdoData->Port2isInput) {
			tmp = READ_PORT_UCHAR(pports[1].Controller + 2);
			tmp |= 0x20;
			WRITE_PORT_UCHAR(pports[1].Controller + 2, (UCHAR)tmp);
		}
	}

	//
	// Check SpindleCW/SpindleCCW, Flood, and Mist signals to avoid
	// checking them in ISR later on. Set Mach signal numbers for them
	// to 0 if they are not valid. This is probably done by Mach itself
	// but it won't hurt to check it here, at Mach startup. This way we
	// will be only checking if they are 0 in ISR instead of that big
	// check if they fall into one of two ranges.
	//
	if (!((mach->SpindleCW >= EXTACT1 && mach->SpindleCW <= EXTACT6) ||
			(mach->SpindleCW >= EXTACT7 && mach->SpindleCW <= EXTACT20)))
		mach->SpindleCW = 0;

	if (!((mach->SpindleCCW >= EXTACT1 && mach->SpindleCCW <= EXTACT6) ||
			(mach->SpindleCCW >= EXTACT7 && mach->SpindleCCW <= EXTACT20)))
		mach->SpindleCCW = 0;

	if (!((mach->Flood >= EXTACT1 && mach->Flood <= EXTACT6) ||
			(mach->Flood >= EXTACT7 && mach->Flood <= EXTACT20)))
		mach->Flood = 0;

	if (!((mach->Mist >= EXTACT1 && mach->Mist <= EXTACT6) ||
			(mach->Mist >= EXTACT7 && mach->Mist <= EXTACT20)))
		mach->Mist = 0;

	FdoData->TmrFreq = 25;	// We start with 25KHz kernel frequency
	MachSetTimerFreq(FdoData);

	//
	// We are not running yet (MACH_RUN not set)
	// and interrupts are disabled. Now get
	// indexes from the board to Mach3 (they are
	// probably all 0 for now but synchronize
	// anyways.)
	//
	for (i = 0; i < 6; i++) {
		if (mach->Axis[i].Enable) {
			mach->Axis[i].Index = FdoData->CPDRegsBase->mach_idx[i];
			mach->StatHolders[i] = mach->StatLast[i] = 0;
			AxData[i].PendingIDX = mach->Axis[i].Index;
		}
	}

	mach->Sync = TRUE;

	//
	// Zero out all encoders. It won't hurt even if they are not enabled
	// in Mach3 config so we are resetting them all without checking.
	//
	dummy = FdoData->CPDRegsBase->mach_enc;
	for (i = 0; i < 7; i++) {
		mach->Encoders[i].Count[0] = mach->Encoders[i].Count[1] = 0;
		mach->Encoders[i].LastStateA = mach->Encoders[i].LastStateB = FALSE;
	}

	//
	// Now clear all interrupts, enable periodic, ADC, DAC,
	// and fault interrupts, set driver state to EStopped
	// and start the show with MACH_RUN. Further actions
	// will be taken by the ISR FSM.
	//
	FdoData->CPDRegsBase->mach_ctl = ESTOP_REQ | FIFO_PURGE |
						MACH_CYCLE_BITS(FdoData->BasePEClockDiv);
	dummy = FdoData->CPDRegsBase->mach_ctl;

	FdoData->DrvState = DrvEStopped;
	mach->EStopReason = CONFIG;

	//
	// Initialize ADC so it can be used in ISR.
	//
	tmp = FdoData->CPDRegsBase->irq_raw;
	tmp |= CP_IRQ_ADC;
	FdoData->CPDRegsBase->irq_raw = tmp;
	do {tmp = FdoData->CPDRegsBase->irq_raw;} while ((tmp & CP_IRQ_ADC) == 0);
	FdoData->CPDRegsBase->adc = ADC_FAST;
	tmp = FdoData->CPDRegsBase->irq_raw;
	tmp |= CP_IRQ_ADC;
	FdoData->CPDRegsBase->irq_raw = tmp;
	do {tmp = FdoData->CPDRegsBase->irq_raw;} while ((tmp & CP_IRQ_ADC) == 0);
	FdoData->CPDRegsBase->adc = ADC_IN0;
	tmp = FdoData->CPDRegsBase->irq_raw;
	FdoData->LastADC = -1;

	//
	// Same for DAC
	//
	tmp = FdoData->CPDRegsBase->irq_raw;
	tmp |= CP_IRQ_DAC;
	FdoData->CPDRegsBase->irq_raw = tmp;
	do {tmp = FdoData->CPDRegsBase->irq_raw;} while ((tmp & CP_IRQ_DAC) == 0);
	FdoData->CPDRegsBase->dac = DAC_B_WORD(FdoData->DACData[1]);
	tmp = FdoData->CPDRegsBase->irq_raw;
	tmp |= CP_IRQ_DAC;
	FdoData->CPDRegsBase->irq_raw = tmp;
	do {tmp = FdoData->CPDRegsBase->irq_raw;} while ((tmp & CP_IRQ_DAC) == 0);
	FdoData->CPDRegsBase->dac = DAC_A_WORD(FdoData->DACData[0]);;
	tmp = FdoData->CPDRegsBase->irq_raw;
	FdoData->LastDAC = 0;

	//
	// Reset ALL interrupts that might be pending so no IRQ generated once
	// we have some of them enabled before finishing this routine. Timer
	// IRQ won't be generated either because of 1 PCI clock delay from
	// MACH_RUN to actually starting the FSM. It is built into the CPLD
	// firmware. There will be no ADC/DAC interrupts because they just
	// loaded with commands and it takes ages for them to get sent via SPI.
	// Spindle IDX/Timing IRQ should not happen either because we are just
	// starting so Spindle should be stopped now.
	//
	// N.B. Fix for 2.0.0.7 -- we do NOT enable DAC/ADC interrupts so they
	//      won't be messing the pulse train. DAC/ADC are polled at each
	//      timer interrupt and updated if ready.
	//
	FdoData->CPDRegsBase->irq_raw = CP_IRQ_ALL_IRQS;
	dummy = FdoData->CPDRegsBase->irq_raw;
	tmp = CP_IRQ_MACH_TMR | CP_IRQ_MACH_FAULT; // | CP_IRQ_ADC | CP_IRQ_DAC;
	if (mach->InSigs[INDEX].Active) tmp |= CP_SPINDLE_IDX;
	if (mach->InSigs[TIMING].Active) tmp |= CP_SPINDLE_TIMING;
	FdoData->IRQMask = tmp;
	FdoData->CPDRegsBase->irq_mask = tmp;
	dummy = FdoData->CPDRegsBase->irq_mask;

	FdoData->Running = TRUE;

	mach->DataInDone = mach->DataOutDone = TRUE;
	FdoData->DataInLowDone = FdoData->DataOutLowDone = 0;

	FdoData->CPDOutBuf = 0;

	tmp = FdoData->CPDRegsBase->mach_ctl;
	tmp |= MACH_RUN;
	FdoData->CPDRegsBase->mach_ctl = tmp;
	dummy = FdoData->CPDRegsBase->mach_ctl;	// Push it through PCI Bus...

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "<--MachDoIoctlStartTimer()\n");
#endif

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "\n-#####################################################-\n");
	if (mach->InSigs[INDEX].Active) {
		DebugPrint(TRACE, DBG_IOCTLS, "+###- INDEX Active%s, Debounce %d\n", mach->InSigs[INDEX].Negated ? "(Negated)" : "", mach->SpinDebounce);
	}
	if (mach->InSigs[TIMING].Active) {
		DebugPrint(TRACE, DBG_IOCTLS, "+###- TIMING Active%s, Debounce %d\n", mach->InSigs[TIMING].Negated ? "(Negated)" : "", mach->Debounce);
	}
	DebugPrint(TRACE, DBG_IOCTLS, "+###- PCIFreq %d\n", FdoData->PCIFreq);
	DebugPrint(TRACE, DBG_IOCTLS, "\n-#####################################################-\n");
#endif

	return TRUE;
}


// *************************************************************************
// * This one changes timer base frequency (25/35/45/60/65/75/100KHz.)     *
// * It is called as SynchCritSection routine hence BOOLEAN (always TRUE.) *
// *************************************************************************
BOOLEAN MachSetTimerFreq(PFDO_DATA FdoData)
{
	UINT32					tmp, div;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "-->MachSetTimerFreq() TmrFreq %d\n",
														FdoData->TmrFreq);
#endif

	switch (FdoData->TmrFreq) {
		//
		// TmrFreq -- Base frequency ("kernel frequency") in KHz.
		//
		case 25:
		case 35:
		case 45:
		case 60:
		case 65:
		case 75:
		case 100:
			div = FdoData->PCIFreq / (FdoData->TmrFreq * 1000) - 1;
			break;

		default:
			return TRUE;
	}

	//
	// Stat period is 250mS -> 250000uS. FdoData->TmrFreq is in KHz.
	// 1 / FdoData->TmrFreq is cycle duration in mS, 1000 / FdoData->TmrFreq
	// is cycle duration in uS. Cycle IRQ is generated every 5th timer clock
	// so IRQ period in uS is 5 * 1000 / FdoData->TmrFreq. Cycles per stat
	// is stat_period / irq_period = 250000 / irq_period =
	//   250000 / (5 * 1000 / FdoData->TmrFreq) =
	//   (250000 * FdoData->TmrFreq) / (5 * 1000) = 50 * FdoData->TmrFreq
	// This is steady value when not Lathe Threading. When threading the
	// base clock is constantly adjusted to match actual Spindle RPM so this
	// value won't be right. That is why we use another value,
	// StatUnitsPerCycle that is proportional to the actual base clock
	// frequency. We add this value to the StatCntr on each and every cycle
	// and check if StatCntr got greater than maximum value, UnitsPerStat.
	// We calculate a BASE value here and use an adjusted one for actual
	// stats. That base value is UnitsPerStat divided by CyclesPerStat above.
	//
	// Jog Velocity is Pulses/Second multiplied by 2048 so we calculate
	// how many Velocity Units is required for Jog to put out a step.
	//
	//FdoData->CyclesPerStat = 50 * FdoData->TmrFreq;
	FdoData->BaseStatUnitsPerCycle = UnitsPerStat / (50 * FdoData->TmrFreq);
	FdoData->CurStatUnitsPerCycle = FdoData->BaseStatUnitsPerCycle;
	FdoData->JogUnitsPerStep = 2048 * FdoData->TmrFreq * 1000;

	FdoData->BasePEClockDiv = div;

	tmp = FdoData->CPDRegsBase->mach_ctl;
	tmp &= ~MACH_CYCLE_MASK;
	tmp |= MACH_CYCLE_BITS(div);
	FdoData->CPDRegsBase->mach_ctl = tmp;
	tmp = FdoData->CPDRegsBase->mach_ctl;	// Push it through PCI...

	mach->PulseCount = FdoData->PCIFreq / FdoData->BasePEClockDiv;

	FdoData->ClockAtBase = TRUE;

#ifdef DEBUG
	DebugPrint(TRACE, DBG_IOCTLS, "<--MachSetTimerFreq() BaseStatUnitsPerCycle %d JogUnitsPerStep %d\n",
				FdoData->BaseStatUnitsPerCycle, FdoData->JogUnitsPerStep);
#endif

	return TRUE;
}


// *************************************************************************
// * Gets all inputs from Port{1,2} if mapped and CENTIPEDE. Those are     *
// * read into appropriate buffers first in BYTEs for Ports and as ULONG   *
// * for CENTIPEDE and then an array of mapping/handling functions moves   *
// * bits to the Mach shared memory block. That array is initialized as    *
// * a part of Init IOCTL according to Mach 3 signal mappings. Ports are   *
// * supposed to be already configured for the proper direction (for their *
// * Data register, offset 0). If a port is configured for INPUT on its    *
// * Data register it is also read as a part of this routine, otherwise    *
// * only Status register is read. Depends on FdoData->PortXisInput var.   *
// *************************************************************************
VOID MachGetInputs(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	pGetSigFunc				*GetSigMap = FdoData->GetSigMap;
	pGetSigFunc				*GetMPGSigMap = FdoData->GetMPGSigMap;
	struct InputInfo		*InSig = mach->InSigs;
	BOOLEAN					*MPGData = (BOOLEAN *)FdoData->MPGData;

	PUINT32					InData = &FdoData->PortsInBuf;
	PUCHAR					Port1Addr = FdoData->PPortInfo[0].Controller;
	PUCHAR					Port2Addr = FdoData->PPortInfo[1].Controller;

	//
	// Get input data from Parallel Ports into FdoData->PortsInBuf
	//
	if (FdoData->Port1On || FdoData->Port2On) {

		FdoData->PortsInBuf = 0;

		if (FdoData->Port2On && Port2Addr != 0) {

			FdoData->PortsInBuf |= READ_PORT_UCHAR(Port2Addr + 1);

			FdoData->PortsInBuf <<= 8;

			if (FdoData->Port2isInput) {

				FdoData->PortsInBuf |= READ_PORT_UCHAR(Port2Addr);
			}

			FdoData->PortsInBuf <<= 8;
		}

		if (FdoData->Port1On && Port1Addr != 0) {

			FdoData->PortsInBuf |= READ_PORT_UCHAR(Port1Addr + 1);

			FdoData->PortsInBuf <<= 8;

			if (FdoData->Port1isInput) {

				FdoData->PortsInBuf |= READ_PORT_UCHAR(Port1Addr);
			}

		} else {

			FdoData->PortsInBuf <<= 8;
		}
	}

	//
	// Get input data from CENTIPEDE.
	//
	FdoData->CPDInBuf = FdoData->CPDRegsBase->din;

	//
	// We are going from the last signal to the first one, i.e backwards.
	// That allows us to use ECX register for indexing thus saving one EDX
	// register addition/incrementing/etc. per signal i.e. 51 commands
	// (2 additional commands out of the loop for adjusting addresses are
	// used so this is 53-2=51.)
	//
	// We can not use this trick on InSigs because INSIG_SIZE is 14 and
	// scaling factor on index register (ECX) can only be 2/4/8.
	//
	__asm {
			mov		ecx, nSigs					; Input signals number

			mov		edx, dword ptr [GetSigMap]	; Call table base addr
			sub		edx, PFUNC_SIZE				; Adjust for indexing with ecx

			mov		edi, dword ptr [InSig]		; InSigs array in MachBlock
			add		edi, (nSigs-1)*INSIG_SIZE	; Last InSigs entry

			mov		eax, dword ptr [InData]		; &FdoData->PortsInBuf

		insigs_loop:
			test	byte ptr [edi], -1			; Active?
			jz		insigs_skip					; Next signal if not Active

			test	dword ptr [edx+ecx*4], -1	; Skip if call table entry
			jz		insigs_skip					; is zero (invalid pin)

			call	dword ptr [edx+ecx*4]		; Process signal

		insigs_skip:
			sub		edi, INSIG_SIZE
			loop	insigs_loop

			; Second part, kinda hacky -- getting MPG inputs.
			; The hack is that 'sub edi, 3' -- InSigxxxx functions
			; are supposed to set InSig[x].Active BOOLEAN that has
			; offset of 8 bytes from the beginning. MPGData is set
			; of BOOLEANs so there is no offset that is why we add 5
			; to point it to the last BOOLEAN and subtract 8 that
			; compensates for [edi+8] in InSigxxxx function.

			mov		ecx, 6						; MPG signals number

			mov		edx, dword ptr [GetMPGSigMap]	; Call table base addr
			sub		edx, PFUNC_SIZE				; Adjust for indexing with ecx

			mov		edi, dword ptr [MPGData]	; FdoData->MPGData
			sub		edi, 3						; Last MPGData entry - 8

		mpgsigs_loop:
			test	dword ptr [edx+ecx*4], -1	; Skip if call table entry
			jz		mpgsigs_skip				; is zero (invalid pin)

			call	dword ptr [edx+ecx*4]		; Process signal

		mpgsigs_skip:
			dec		edi							; MPGData are byte long
			loop	mpgsigs_loop
	}
}


// ************************************************************************
// * Similar to GetInputs above but for propagating Mach 3 output signals *
// * to the actual hardware outputs. Actual data is constructed in Port   *
// * and CENTIPEDE buffers that are then output to the actual hardware in *
// * 1 or 2 bytes per parallel port (depending on configuration) and one  *
// * 32-bit value for CENTIPEDE.                                          *
// ************************************************************************
VOID MachPutOutputs(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	pPutSigFunc				*PutSigMap = FdoData->PutSigMap;
	struct OutputInfo		*OutSig = mach->OutSigs;
	PUINT32					OutData = &FdoData->PortsOutBuf;
	PUCHAR					Port1Addr = FdoData->PPortInfo[0].Controller;
	PUCHAR					Port2Addr = FdoData->PPortInfo[1].Controller;
	UINT32					dummy;

	//
	// This stuffs all Mach3 output signals into FdoData->PortsOutBuf and
	// FdoData->CPDOutBuf that will be sent to actual ports in single write
	// per port.
	// We also do these from last to first using ECX register for indexing
	// like it is done in MachGetInputs().
	//
	__asm {
			mov		edi, dword ptr [OutData]	; edi <- &FdoData->PortsOutBuf
			and		dword ptr [edi], 0			; PortsOutBuf <- 0
;			and		dword ptr [edi+4], 0		; CPDOutBuf <- 0
			mov		ecx, nSigsOut				; Output signals number

			mov		edx, dword ptr [PutSigMap]	; Call table base addr
			sub		edx, PFUNC_SIZE				; Adjust for indexing with ecx

			mov		esi, dword ptr [OutSig]		; OutSigs array in MachBlock
			add		esi, (nSigsOut-1)*OUTSIG_SIZE	; Last OutSigs entry

		outsigs_loop:
			test	byte ptr [esi], -1			; active?
			jz		outsigs_skip				; Next signal if not active

			test	dword ptr [edx+ecx*4], -1	; Skip if call table
			jz		outsigs_skip				; entry is zero (invalid pin)

			mov 	eax, [esi+3]				; al <- Negated, ah <- Activated
			call	dword ptr [edx+ecx*4]		; Process signal

		outsigs_skip:
			sub		esi, OUTSIG_SIZE
			loop	outsigs_loop
	}

	//
	// Send FdoData->PortsOutBuf to actual Parallel Ports.
	//
	if (FdoData->Port1On && Port1Addr != 0) {

		if (!FdoData->Port1isInput) {

			WRITE_PORT_UCHAR(Port1Addr, (UCHAR)FdoData->PortsOutBuf);
		}

		FdoData->PortsOutBuf >>= 8;
		if (FdoData->Port1isInput) FdoData->PortsOutBuf |= 0x20;	// Dir bit
		WRITE_PORT_UCHAR(Port1Addr + 2, (UCHAR)FdoData->PortsOutBuf);
		FdoData->PortsOutBuf >>= 8;

	} else {

		FdoData->PortsOutBuf >>= 16;
	}

	if (FdoData->Port2On && Port2Addr != 0) {

		if (!FdoData->Port2isInput) {

			WRITE_PORT_UCHAR(Port2Addr, (UCHAR)FdoData->PortsOutBuf);
		}

		FdoData->PortsOutBuf >>= 8;
		if (FdoData->Port2isInput) FdoData->PortsOutBuf |= 0x20;	// Dir bit
		WRITE_PORT_UCHAR(Port2Addr + 2, (UCHAR)FdoData->PortsOutBuf);
	}

	//
	// Send FdoData->CPDOutBuf to CENTIPEDE.
	//
	FdoData->CPDRegsBase->dout = FdoData->CPDOutBuf;
	dummy = FdoData->CPDRegsBase->dout;
}


// ****************************************************
// * This one stops spindle and cooling in emergency. *
// *****************************************************
VOID MachStopSpindle(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	UINT32					tmp;

	//
	// This stops Dir/Step Spindle Pulsing immediately if we use it.
	// It is probably better way to stop it in emergency than just
	// decelerate it to stop... It is emergency so ALL motion must be
	// stopped IMMEDIATELY, right?
	//
	if (mach->Axis[SPINDLE].Enable) {
		mach->Axis[SPINDLE].MaxVelocity = mach->Axis[SPINDLE].CurVelocity = 0;
		mach->Axis[SPINDLE].Jogging = FALSE;
		tmp = FdoData->CPDRegsBase->mach_sp_ctl;
		tmp &= ~SP_ST_RUN;
		FdoData->CPDRegsBase->mach_sp_ctl = tmp;
		tmp = FdoData->CPDRegsBase->mach_sp_ctl;
	}

	//
	// Now SpindleCW/SpindleCCW and Flood/Mist signals...
	//
	if (mach->SpindleCW != 0) mach->OutSigs[mach->SpindleCW].Activated = FALSE;
	if (mach->SpindleCCW != 0) mach->OutSigs[mach->SpindleCCW].Activated = FALSE;
	if (mach->Flood != 0) mach->OutSigs[mach->Flood].Activated = FALSE;
	if (mach->Mist != 0) mach->OutSigs[mach->Mist].Activated = FALSE;

	MachPutOutputs(FdoData);
}


// *******************************************************
// * This one tries to get the board out of EStop state. *
// * Returns STATUS_SUCCESS if successful, otherwise an  *
// * appropriate Mach3 EStopReason code is returned.     *
// *******************************************************
NTSTATUS MachLeaveEStop(PFDO_DATA FdoData)
{
	volatile PCPD_REGS		brd = FdoData->CPDRegsBase;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	UINT32					tmp;

	//
	// Try to reset limit switches hit bits and EStop.
	//
	tmp = brd->mach_ctl;
	if (mach->LimitsOff)
		tmp |= NO_LIMITS;
	else
		tmp &= ~NO_LIMITS;
	brd->mach_ctl = tmp;
	tmp = brd->mach_ctl;

	brd->mach_stat = ST_LIMIT_HIT | ST_ESTOP;
	tmp = brd->mach_stat;

	//
	// Try to reset all emergency interrupts. That will only succeed
	// if the original condition that caused them is gone.
	//
	brd->irq_raw = CP_IRQ_MACH_FAULT | CP_IRQ_MACH_FOVF;

	//
	// Tell board to run
	//
	tmp = brd->mach_ctl;
	tmp &= ~ESTOP_REQ;
	tmp |= MACH_RUN;
	brd->mach_ctl = tmp;

	//
	// Check if it out of EStop and running OK.
	//
	tmp = brd->mach_stat;
	if ((tmp & (ST_ESTOP | ST_BUSY)) != 0) {
		if ((tmp & ST_LIMIT_HIT) != 0)
			return LIMITSWITCH;
		else
			return ESTOPINPUTLINE;
	}

	return STATUS_SUCCESS;
}


// **************************************************************************
// * This stuffs Mach3 Planned Motion entries in CENTIPEDE FIFO on every    *
// * timer interrupt. Each entry makes 5 FIFO items. Caller checks if there *
// * is something in the Mach3 queue so it is not necessary to check this   *
// * here -- we are not called if Mach3 queue is empty. Everything else is  *
// * done here.                                                             *
// **************************************************************************
NTSTATUS MachPlannedMotion(PFDO_DATA FdoData)
{
	volatile PUINT32		pFifo = &FdoData->CPDRegsBase->mach_fifo;
	INT						FifoRoom;
	INT						i, j;
	PINT					CurPoints;
	PCHAR					CurShifter;
	static UINT32			CurDirs = 0;
	CHAR					Shifter[6];
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	volatile PAXIS_DATA		AxData = FdoData->AxisData;
	PINT					LimitsPos = mach->SoftLimitsHi;
	PINT					LimitsNeg = mach->SoftLimitsLo;
	BOOLEAN					OutOfLimits = FALSE;

	if (!FdoData->Running) return STATUS_DEVICE_NOT_READY;

	// Not necessary but won't hurt...
	if (mach->TrajIndex == (mach->TrajHead & 0xfff)) return STATUS_SUCCESS;

	if ((FifoRoom = FIFO_FREE) == 0) return STATUS_SUCCESS;

	for (i = 0; i < FifoRoom; i++) {

		CurPoints = mach->Trajectories[mach->TrajIndex].Points;
		CurShifter = mach->Trajectories[mach->TrajIndex].shifter;

		*(PULONG)Shifter = 0;
		*(PUSHORT)&Shifter[4] = 0;

#ifdef DEBUG
	if (mach->Trajectories[mach->TrajIndex].Command != 0) {
		DebugPrint(TRACE, DBG_PLANNED_MOTION, "-#####################################################-\n");
		DebugPrint(TRACE, DBG_PLANNED_MOTION, "+###- Trajectory Command %d, ID %d\n", mach->Trajectories[mach->TrajIndex].Command, mach->Trajectories[mach->TrajIndex].ID);
		DebugPrint(TRACE, DBG_PLANNED_MOTION, "+###- Points  %d %d %d %d %d %d\n",
			mach->Trajectories[mach->TrajIndex].Points[0],
			mach->Trajectories[mach->TrajIndex].Points[1],
			mach->Trajectories[mach->TrajIndex].Points[2],
			mach->Trajectories[mach->TrajIndex].Points[3],
			mach->Trajectories[mach->TrajIndex].Points[4],
			mach->Trajectories[mach->TrajIndex].Points[5]);
		DebugPrint(TRACE, DBG_PLANNED_MOTION, "+###- Shifter  %02d %02d %02d %02d %02d %02d\n",
			mach->Trajectories[mach->TrajIndex].shifter[0],
			mach->Trajectories[mach->TrajIndex].shifter[1],
			mach->Trajectories[mach->TrajIndex].shifter[2],
			mach->Trajectories[mach->TrajIndex].shifter[3],
			mach->Trajectories[mach->TrajIndex].shifter[4],
			mach->Trajectories[mach->TrajIndex].shifter[5]);
		DebugPrint(TRACE, DBG_PLANNED_MOTION, "-#####################################################-\n");
	}
#endif

		for (j = 0; j < 6; j++) {

			if (!mach->Axis[j].Enable) continue;
			if (AxData[j].Type == Slave) continue;

			//
			// 0 is move to PLUS direction.
			// We KEEP the last Dir if no step. If there is a step
			// to be taken we set the new Dir to the step direction.
			// This should not be necessary but some stepper drivers
			// don't like Dir line flipping with no step.
			//
			if (CurPoints[j] > AxData[j].PendingIDX)
				CurDirs &= ~(1 << (j * 2));
			if (CurPoints[j] < AxData[j].PendingIDX)
				CurDirs |= 1 << (j * 2);

			// We KEEP CurDirs bit if no step...

			Shifter[j] = CurShifter[j];

			AxData[j].PendingIDX = CurPoints[j];

			if (AxData[j].Type == Master) {
				Shifter[AxData[j].Slave] = CurShifter[j];
				if (CurPoints[j] > AxData[j].PendingIDX)
					CurDirs &= ~(1 << (AxData[j].Slave * 2));
				if (CurPoints[j] < AxData[j].PendingIDX)
					CurDirs |= 1 << (AxData[j].Slave * 2);
				AxData[AxData[j].Slave].PendingIDX = CurPoints[j];
			}

			if (!mach->SoftLimitEnable) continue;

			if (AxData[j].PendingIDX >= LimitsPos[j] ||
					AxData[j].PendingIDX <= LimitsNeg[j]) {

					OutOfLimits = TRUE;
			}
		}

		//
		// 6x5 Mach points to 5x6 fifo_entries matrix transposition
		// with direction bits insertion and actual output to the board.
		//
		__asm {

			mov		ecx, dword ptr [Shifter]
			mov		ax, cx
			shr		ecx, 16
			mov		dx, word ptr [Shifter+4]
			mov		ebx, dword ptr [pFifo]

			MK_MOVE
			MK_MOVE
			MK_MOVE
			MK_MOVE
			MK_MOVE
		}

		mach->TrajIndex++;
		mach->TrajIndex &= 0xfff;

		if (OutOfLimits) return SOFTLIMITS;

		if (mach->TrajIndex == (mach->TrajHead & 0xfff)) break;
	}

	return STATUS_SUCCESS;
}


// ***************************************
// * This is a JOG/Probe/Homing routine. *
// ***************************************
NTSTATUS MachJog(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT		mach = FdoData->MachSharedBlock;
	volatile struct AxisInfo	*pAx, *pAxSlave;
	volatile PAXIS_DATA			AxData = FdoData->AxisData;
	INT							Dir;
	INT							TargetVelocity;
	INT							FifoRoom;
	UINT32						FifoData[5], tmp;
	CHAR						MoreWorkRequired;
	INT							i, j, dummy;

	//
	// FIFO full, wait for it to drain somewhat.
	//
	if ((FifoRoom = FIFO_FREE) == 0) return STATUS_MORE_PROCESSING_REQUIRED;

	if (FdoData->DrvState == DrvHomingPrePass ||
					FdoData->DrvState == DrvHomingClearSwitches)
		tmp = FdoData->CPDRegsBase->din;

	do {
		//
		// Zero out our 5 step deep motion buffer. We work in 5 step
		// increments because we have timer IRQ every 5 CENTIPEDE timer
		// clocks. It is this way because Planned Motion is done in 5-cycle
		// steps (see TrajPoint shifter[6] member) so the hardware is
		// designed to support this.
		//
		for (i = 0; i < 5; i++) FifoData[i] = 0;

		MoreWorkRequired = 0;

		//
		// For all 6 axes...
		//
		for (i = 0; i < 6; i++) {

			pAx = &mach->Axis[i];

			//
			// Axis is not enabled, skip it.
			//
			if (!pAx->Enable) continue;

			//
			// Slave axis is independent when Dereferencing to square a pair.
			// In all other cases it mirrors whatever is done by its Master
			// so we skip it here and all the motion data for this axis will
			// be generated and stuffed as a part of its Master processing.
			//
			if (AxData[i].Type == Slave && FdoData->DrvState != DrvHomingDeref)
				continue;

			//
			// For HomingPrePass and HomingClearSwitches we only work on
			// those axes that were marked with Hit by the ISR.
			//
			if ((FdoData->DrvState == DrvHomingPrePass && !AxData[i].Hit) ||
				(FdoData->DrvState == DrvHomingClearSwitches && !pAx->DeRef))
				continue;

			//
			// Doing nothing on this axis, skip it.
			// There should be no reason for checking Probing and Homing,
			// they should be set IN ADDITION to Jogging to differentiate
			// a regular jog from probing/homing one...
			//
			if (!pAx->Jogging) continue;

			//
			// Get Slave axis pointer if any.
			//
			pAxSlave = NULL;

			if (AxData[i].Type == Master)
				pAxSlave = &mach->Axis[AxData[i].Slave];

			//
			// Axis stopped, nothing to do, skip.
			//
			if (pAx->Dec && pAx->CurVelocity == 0) continue;

			//
			// We are checking Home Switches here for HomingPrePass and
			// HomingClearSwitches (the very last move while Homing) pass,
			// they are NOT setup to generate interrupt and we don't care
			// where axes stop as far as they went off switches. We set
			// Dec here when they are off switches that will stop those
			// axes somewhere. ClearSwitchesMask is set by the ISR.
			//
			if (FdoData->DrvState == DrvHomingPrePass ||
					FdoData->DrvState == DrvHomingClearSwitches) {
				if ((tmp & AxData[i].ClearSwitchesMask) == 0) {
					pAx->Dec = TRUE;
					//pAx->Acc = FALSE;
					if (pAxSlave) {
						pAxSlave->Dec = TRUE;
						//pAxSlave->Acc = FALSE;
					}
				}
			}

			//
			// For other routines, to save time on checking which
			// axis is currently active
			//
			AxData[i].Active = TRUE;
			if (pAxSlave) AxData[AxData[i].Slave].Active = TRUE;

			//
			// 5 steps deep FIFO stage, that is done inside "For all Axes"
			// loop so it will generate 5 32-bit values to be loaded in the
			// board FIFO.
			//
			for (j = 0; j < 5; j++) {

				//
				// Direction bits for the axis and its slave (if required.)
				// We take JoggDir as is -- it is setup to a proper
				// direction by the ISR FSM when changing states.
				//
				Dir = (pAx->JoggDir == 0) ? 0 : 1;

				FifoData[j] |= Dir << (i * 2);
				if (pAxSlave && FdoData->DrvState != DrvHomingDeref)
					FifoData[j] |= Dir << (AxData[i].Slave * 2);

				//
				// We are decelerating... We calculate target velocity only
				// at this stage, real step signals will be generated later
				// based on our running velocity.
				//
				if (pAx->Dec) {

					//
					// Already stopped, no more work, break out...
					//
					if (pAx->CurVelocity == 0) {
						MoreWorkRequired &= ~(1 << i);
						break;
					}

					//
					// Slow down one Acceleration step, clamp to zero.
					//
					pAx->CurVelocity -= pAx->Acceleration;
					if (pAx->CurVelocity < 0) pAx->CurVelocity = 0;

					//
					// Decelerated to stop, no more work, break out...
					//
					if (pAx->CurVelocity == 0) {
						MoreWorkRequired &= ~(1 << i);
						break;
					}

					//
					// Velocity is still not zero, more steps to do for this
					// axis (MoreWorkRequired is bit-per-axis so it will only
					// become zero when ALL the jogging axes are stopped.)
					//
					MoreWorkRequired |= 1 << i;

				//
				// Accelerating or keeping moving...
				//
				} else {

					//
					// Clamp velocity to absolute maximum
					//
					TargetVelocity = (pAx->MaxVelocity > pAx->MasterVelocity) ?
						pAx->MasterVelocity : pAx->MaxVelocity;

					//
					// Speed up
					//
					if (pAx->CurVelocity < TargetVelocity) {

						pAx->CurVelocity += pAx->Acceleration;
						if (pAx->CurVelocity > TargetVelocity)
							pAx->CurVelocity = TargetVelocity;

					//
					// Slow down (target velocity can be overridden by
					// the user via Mach3.) Don't go below the new target,
					// clamp to it if deceleration step made us slower.
					//
					} else if (pAx->CurVelocity > TargetVelocity) {

						pAx->CurVelocity -= pAx->Acceleration;
						if (pAx->CurVelocity < TargetVelocity)
							pAx->CurVelocity = TargetVelocity;
					}

					//
					// Set bit. We are doing motion so more work is needed
					// until we come to full stop.
					//
					MoreWorkRequired |= 1 << i;
				}

				//
				// Copy target velocity to a slave axis if there is one
				// and we are not derefencing when all axes are independent.
				//
				if (pAxSlave && FdoData->DrvState != DrvHomingDeref)
					pAxSlave->CurVelocity = pAx->CurVelocity;

				//
				// 1 step is taken for every 0x800 units. Velocity is number
				// of units per minimum step interval (1 / kernel frequency.)
				// Each entry we make here is one minimum step interval ergo
				// velocity is number of units to be added to the Quantum
				// Counter. Every time counter goes over 0x7ff a step is
				// made and counter wraps around.
				//
				AxData[i].QuantCntr += pAx->CurVelocity;

				if (AxData[i].QuantCntr > FdoData->JogUnitsPerStep) {

					//
					// Wrap around...
					//
					AxData[i].QuantCntr -= FdoData->JogUnitsPerStep;

					//
					// Update PendingIDX with a position it is supposed to be
					// after this step is done. It is NOT an actual running
					// position because those steps can be still sitting in
					// board FIFO. We use those Pending indexes to set axes
					// to decelerate at the set TripPoints (to stop at some
					// precalculated positions when Probing/Homing.)
					//
					// We do a little cheating here when Dereferencing...
					// Sometimes we might overshoot home position by ONE
					// step. To avoid this we will SKIP that step if it
					// going to happen. Hopefully nobody will notice it :)
					// We were decelerating, motion is very slow and were
					// about to stop anyways...
					//
					if (!(FdoData->DrvState == DrvHomingDeref && pAx->DeRef &&
								AxData[i].PendingIDX == AxData[i].HitIDX)) {
						if (Dir == 1)
							AxData[i].PendingIDX--;
						else
							AxData[i].PendingIDX++;

						FifoData[j] |= 1 << (i * 2 + 1);
					}

					//
					// Do the same for a slave axis if any and state is right.
					//
					if (pAxSlave && FdoData->DrvState != DrvHomingDeref) {
						if (Dir == 1)
							AxData[AxData[i].Slave].PendingIDX--;
						else
							AxData[AxData[i].Slave].PendingIDX++;
						FifoData[j] |= 1 << (AxData[i].Slave * 2 + 1);
					}

					//
					// Now check if the new position will hit TripPoint and
					// do necessary actions if it will.
					//
					if (FdoData->DrvState == DrvHomingDeref) {
						//
						// No slave axes when dereferencing...
						//
						if (AxData[i].PendingIDX == AxData[i].TripPoint) {
							//
							// We delay switching to deceleration if
							// DeRefAdjust is set. It is set in DeRef prep
							// routine if it knows we are going to undershoot
							// 1 step. We set DoAdjust here so we switch
							// later on before the next step is taken.
							//
							if (AxData[i].DeRefAdjust) {
									AxData[i].DoAdjust = TRUE;
									AxData[i].DeRefAdjust = FALSE;
							} else {
								pAx->Dec = TRUE;
								//pAx->Acc = FALSE;
							}
						}

					} else if (FdoData->DrvState == DrvProbing) {

						if (AxData[i].PendingIDX == AxData[i].TripPoint) {
							pAx->Dec = TRUE;
							//pAx->Acc = FALSE;
							if (pAxSlave) {
								pAxSlave->Dec = TRUE;
								//pAxSlave->Acc = FALSE;
							}
						}
					}
				}

				//
				// Delayed switch to deceleration just before the next step
				// would've been taken on next quantum.
				//
				if (AxData[i].DoAdjust) {
					dummy = FdoData->JogUnitsPerStep - (pAx->CurVelocity + pAx->Acceleration);
					if (AxData[i].QuantCntr >= dummy) {
							pAx->Dec = TRUE;
							//pAx->Acc = FALSE;
							AxData[i].DoAdjust = FALSE;
					}
				}
			}

			//if (pAx->Dec) continue;

			pAx->AtSpeed = (pAx->CurVelocity == TargetVelocity) ?
				TRUE : FALSE;
			if (pAxSlave && FdoData->DrvState != DrvHomingDeref)
				pAxSlave->AtSpeed = pAx->AtSpeed;
		}

		for (i = 0; i < 5; i++)
			FdoData->CPDRegsBase->mach_fifo = FifoData[i];

	} while (--FifoRoom > 0 && MoreWorkRequired != 0);

	return (MoreWorkRequired == 0) ? STATUS_SUCCESS : STATUS_MORE_PROCESSING_REQUIRED;
}


// ************************************************************************
// * This is called when we are Probing, Probe Switch is enabled and hit, *
// * i.e. Probe IRQ is detected by the ISR. IDX registers are frozen at   *
// * the ProbeHit position so this routine transfers hit position indexes *
// * to Mach3 then ISR must disable ProbeSwitch and reset the IRQ to      *
// * unfreeze the IDX registers.                                          *
// ************************************************************************
VOID MachHandleProbeHit(PFDO_DATA FdoData)
{
	int							i;
	volatile PMACH_STRUCT		mach = FdoData->MachSharedBlock;
	volatile struct AxisInfo	*pAx, *pAxSlave = NULL;
	volatile PINT32				Idx = FdoData->CPDRegsBase->mach_idx;

	for (i = 0; i < 6; i++) {

		pAx = &mach->Axis[i];

		if (FdoData->AxisData[i].Type == Master)
			pAxSlave = &mach->Axis[FdoData->AxisData[i].Slave];

		if (!pAx->Enable) continue;
		if (!pAx->Probing) continue;
		if (pAx->Slave) continue;

		//
		// Get the latched hit point from the board to DepthCount.
		// IDX registers still keep their latched value until Probe
		// IRQ is reset and Probe Switch disabled. That should be done
		// from the ISR to avoid race conditions (handle all that from
		// a single place, the ISR.)
		//
		pAx->DepthCount = Idx[i];
		pAx->Dec = TRUE;
		//pAx->Acc = FALSE;

		if (pAxSlave) {
			pAxSlave->DepthCount = Idx[i];
			pAxSlave->Dec = TRUE;
			//pAxSlave->Acc = FALSE;
		}
	}
}


// ***********************************************************************
// * This is called when one or more home switches hit when homing.      *
// * Hit point indexes are stored in AxisData structs and corresponding  *
// * axes are set to Decelerate thus stopping them. For axis pairs       *
// * (Master/Slave) they are both set to Decelerate at the same time     *
// * when they both have their switches hit. If only one switch is hit   *
// * it is noted in corresponding AxisData[x] structure and we continue  *
// * to jog until the other one hit. Hit positions are saved so we will  *
// * be able to calculate proper dereference jogs to square all axes.    *
// ***********************************************************************
VOID MachHandleHomeHit(PFDO_DATA FdoData)
{
	INT						i;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	PAXIS_DATA				AxData = FdoData->AxisData;
	volatile PINT32			Idx = FdoData->CPDRegsBase->mach_idx;
	UINT32					SwHit;

	//
	// Read board status to find out which switches were hit. Home switches
	// hit bits are ((i*3)+2) bits where i is axis number.
	//
	SwHit = FdoData->CPDRegsBase->mach_stat;

	for (i = 0; i < 6; i++) {

		//
		// Skip inactive axes and those which switch has not been hit.
		// Set status if active axis switch has not been hit.
		//
		if (!AxData[i].Active) continue;
		if ((SwHit & (1 << ((i * 3) + 2))) == 0) continue;

		//
		// Mark axis as hit and store IDX where it hit the switch. It will be
		// used for calculating dereference move and final "referenced" index.
		//
		AxData[i].Hit = TRUE;
		AxData[i].HitIDX = Idx[i];
		AxData[i].ClearSwitchesMask |= 1 << (i*3+2);

		//
		// Set Dec on the axis if it is a single one or on both paired axes
		// (Master/Slave) if they BOTH were hit. Do NOT set Dec on any of
		// Master/Slave if the other one was not hit. We set a Hit variable
		// and continue jogging until the other one is also hit then we set
		// Dec on both of them at the same time and stop them. HitIDX the
		// used to calculate TripPoints for each axis. Paired axes are setup
		// in such a way that they will stop at the same point thus squaring
		// them. All axes are independent on dereference jog i.e. pairs are
		// "unlocked" that is necessary for squaring.
		//
		switch (AxData[i].Type) {

			case Single:
				mach->Axis[i].Dec = TRUE;
				//mach->Axis[i].Acc = FALSE;
				break;
			case Master:
				if (AxData[AxData[i].Slave].Hit) {
					mach->Axis[i].Dec = TRUE;
					//mach->Axis[i].Acc = FALSE;
					mach->Axis[AxData[i].Slave].Dec = TRUE;
					//mach->Axis[AxData[i].Slave].Acc = FALSE;
				}
				AxData[i].ClearSwitchesMask |= 1 << (AxData[i].Slave*3+2);
				AxData[AxData[i].Slave].ClearSwitchesMask =
										AxData[i].ClearSwitchesMask;
				break;
			case Slave:
				if (AxData[AxData[i].Master].Hit) {
					mach->Axis[i].Dec = TRUE;
					//mach->Axis[i].Acc = FALSE;
					mach->Axis[AxData[i].Master].Dec = TRUE;
					//mach->Axis[AxData[i].Master].Acc = FALSE;
				}
				AxData[i].ClearSwitchesMask |= 1 << (AxData[i].Master*3+2);
				AxData[AxData[i].Master].ClearSwitchesMask =
										AxData[i].ClearSwitchesMask;
		}
	}
}


// **********************************************************************
// * This one prepares all axes for regular Homing jog after they were  *
// * moved off the Home switches in Pre pass.                           *
// *                                                                    *
// * Mach3 sets up DeReference jog in opposite direction when it finds  *
// * the Home switch already activated. That is supposed to catch zero  *
// * when going OFF the switch. This is only done if HomeSecurity is    *
// * is not set in Mach3 config otherwise it will refuse to do anything *
// * until all axes are moved off the Home switches manually.           *
// *                                                                    *
// * However it is NOT a good idea. First of all, there is no guarantee *
// * that switch closes and opens at exactly the same position. So it   *
// * makes sense to always approach them from the same direction. Then  *
// * there is a problem with slave axes -- what if they are racked and  *
// * only one is on the switch? In that case they should either jog in  *
// * opposite directions or one of them will miss its switch on the     *
// * first step because it is ALREADY off the switch and we will have   *
// * no hit point to calculate the final dereference jog to square      *
// * them. We can get their exact locations wrt their Home switches at  *
// * the point they stopped, one on forward move another on reverse one *
// * but that will require an additional squaring move that makes it    *
// * more complicated.                                                  *
// *                                                                    *
// * So we are moving all the axes OFF their home switches first and    *
// * only then starting regular Homing process. We can do it because    *
// * we have separate Home switches on each axis.                       *
// *                                                                    *
// * Axes that were moved in reverse because of being ON Home switch    *
// * marked as Hit by the ISR FSM prior to calling this routine. We use *
// * it to setup a Pre jog. For Master/Slave pairs BOTH axes were moved *
// * in reverse if either one was hit.                                  *
// **********************************************************************
VOID MachPrepareHome1stPass(PFDO_DATA FdoData)
{
	INT						i;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	PAXIS_DATA				AxData = FdoData->AxisData;

	for (i = 0; i < 6; i++) {

		if (AxData[i].Type == Slave) continue;
		if (!mach->Axis[i].Enable) continue;

		if (AxData[i].Hit) {
			AxData[i].Hit = FALSE;
			mach->Axis[i].JoggDir = (mach->Axis[i].JoggDir == 0) ? 1 : 0;
			mach->Axis[i].Homing = TRUE;

			if (AxData[i].Type == Master) {
				AxData[AxData[i].Slave].Hit = FALSE;
				mach->Axis[AxData[i].Slave].JoggDir = mach->Axis[i].JoggDir;
				mach->Axis[AxData[i].Slave].Homing = TRUE;
			}
		}

		if (mach->Axis[i].Homing) {
			mach->Axis[i].Dec = FALSE;
			mach->Axis[i].DeRef = FALSE;
			mach->Axis[i].AtSpeed = FALSE;
			mach->Axis[i].Jogging = TRUE;
			mach->Axis[i].CurVelocity = 0;
			AxData[i].PendingIDX = FdoData->CPDRegsBase->mach_idx[i];
			AxData[i].QuantCntr = 0;
			AxData[i].ClearSwitchesMask = 0;

			if (AxData[i].Type == Master) {
				mach->Axis[AxData[i].Slave].Dec = FALSE;
				mach->Axis[AxData[i].Slave].DeRef = FALSE;
				mach->Axis[AxData[i].Slave].AtSpeed = FALSE;
				mach->Axis[AxData[i].Slave].Jogging = TRUE;
				mach->Axis[AxData[i].Slave].CurVelocity = 0;
				AxData[AxData[i].Slave].PendingIDX =
					FdoData->CPDRegsBase->mach_idx[AxData[i].Slave];
				AxData[AxData[i].Slave].QuantCntr = 0;
				AxData[AxData[i].Slave].ClearSwitchesMask = 0;
			}
		}
	}
}


// ***********************************************************************
// * This one calculates and prepares Dereferencing Jogs for moving axes *
// * off of the Home switches. Locked (Master/Slave) pairs will move     *
// * independently thus squaring them. Very simple routine for now...    *
// ***********************************************************************
VOID MachPrepareDeref(PFDO_DATA FdoData)
{
	INT						i, tmp;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	PAXIS_DATA				AxData = FdoData->AxisData;

	for (i = 0; i < 6; i++) {
		//
		// All axes that are being homed are programmed to stop exactly on
		// Home Switch hit points. That is done by setting a TripPoint
		// midway between HitIDX (home switch hit point index) and current
		// CENTIPEDE IDX (that is where we stopped after overtravelling the
		// home switch.) We use integer division so it might end up 1 step
		// farther off the switch. That means this must be checked by
		// comparing HitIDX and current CENTIPEDE IDX at the final step and
		// the final resulting position should be adjusted by 1 if needed.
		//
		// We set DeRefAdjust if the distance between current position and
		// HitIDX is even. In such a case our dereference jog would almost
		// certainly undershoot by 1 step so this signals jog routine to
		// switch to deceleration slightly later so we will make that missing
		// step. If we overcompensated and an extra step could've resulted
		// that step will be skipped by little cheatcode in jog routine.
		//
		// We also reverse directions here so it moves back off switches.
		//
		tmp = FdoData->CPDRegsBase->mach_idx[i];

		AxData[i].TripPoint = AxData[i].HitIDX + (tmp - AxData[i].HitIDX) / 2;

		AxData[i].DeRefAdjust = (((tmp - AxData[i].HitIDX) % 2) == 0) ? TRUE : FALSE;

		if (AxData[i].Active) {
			mach->Axis[i].DeRef = TRUE;
			mach->Axis[i].Dec = mach->Axis[i].AtSpeed = FALSE;
			mach->Axis[i].Jogging = TRUE;
			mach->Axis[i].Homing = FALSE;
			mach->Axis[i].JoggDir = (mach->Axis[i].JoggDir == 0) ? 1 : 0;
// KSI: Is it us who set those or Mach3 signalling us to use slower speed?
//            mach->HomeSpeed[i] = TRUE;
			AxData[i].Hit = FALSE;
			AxData[i].DoAdjust = FALSE;
			AxData[i].PendingIDX = tmp;
			AxData[i].QuantCntr = 0;
		}
	}
}


// ********************************************************************
// * This is almost final stage of homing process. It is called after *
// * Dereference move is complete and the machine is stopped at Home  *
// * Switches Zero position. It loads proper indexes to CENTIPEDE and *
// * Mach3 and prepare everything required for the final step, moving *
// * off of Home Switches.                                            *
// ********************************************************************
NTSTATUS MachPrepareClearSwitches(PFDO_DATA FdoData)
{
	INT						i, RefIDX;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	PAXIS_DATA				AxData = FdoData->AxisData;
	UINT32					tmp;

	tmp = FdoData->CPDRegsBase->din;

	if ((tmp & ST_HOME_HIT) == 0) return STATUS_SUCCESS;

	for (i = 0; i < 6; i++) {

		if (!mach->Axis[i].DeRef) continue;

		//
		// This is to adjust for ovetravel if we did not stop exactly
		// on the switch hit point position. Not probably necessary with
		// that cheatcode in jog routine but won't hurt so it stays for now.
		//
		RefIDX = FdoData->CPDRegsBase->mach_idx[i] - AxData[i].HitIDX;

		//
		// Use the preset value if AutoZero is set for this axis, 0 if not.
		//
		if (mach->AutoZero[i]) RefIDX += mach->RefCounts[i];

		//
		// Set all Indexes to a configured zero point and use those
		// from now on. We won't stop on exact zeroes but everything
		// be referenced and squared properly when we do.
		//
		FdoData->CPDRegsBase->mach_idx[i] = RefIDX;
		mach->Axis[i].Index = FdoData->CPDRegsBase->mach_idx[i];

		//
		// DeRef is set, we set Jogging so we'll move off of switch
		// at the last step that is coming.
		//
		mach->Axis[i].Jogging = TRUE;
		mach->Axis[i].Dec = FALSE;
	}

	return ((tmp & ST_HOME_HIT) == 0) ? STATUS_SUCCESS : STATUS_MORE_PROCESSING_REQUIRED;
}


// ******************************************************************
// * This is the final stage of homing process. It is called after  *
// * Dereference move is complete and the machine is stopped at its *
// * final homed position. It cleans up all jogs remnants and tells *
// * Mach3 to Sync to new Indexes that will make it reset Homing by *
// * itself after it synced. After that Mach3 is ready to go.       *
// ******************************************************************
VOID MachCompleteHome(PFDO_DATA FdoData)
{
	INT						i;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	PAXIS_DATA				AxData = FdoData->AxisData;

	for (i = 0; i < 6; i++) {

		if (!mach->Axis[i].DeRef) continue;

		mach->Axis[i].Index = FdoData->CPDRegsBase->mach_idx[i];

		mach->Axis[i].DeRef = FALSE;
		mach->Axis[i].Homing = FALSE;
		mach->Axis[i].Probing = FALSE;
		mach->Axis[i].Jogging = FALSE;
		mach->Axis[i].Dec = FALSE;
		mach->Referenced[i] = TRUE;
		AxData[i].Active = FALSE;
		AxData[i].Hit = FALSE;
		AxData[i].DeRefAdjust = FALSE;
		AxData[i].DoAdjust = FALSE;
		AxData[i].QuantCntr = 0;
		AxData[i].ClearSwitchesMask = 0;
	}
}


// **********************************************************************
// * This gets actual Index counts from the board and puts them into    *
// * Axis[x]. Also updates StatsLast[x] in Mach3 structure every 250mS  *
// * with cumulative per-axis number of steps actually taken so it will *
// * be able to display motion speed on every axis. This must be called *
// * every timer tick.                                                  *
// **********************************************************************
VOID MachUpdateIndexes(PFDO_DATA FdoData)
{
	INT						i, tmp;
	BOOLEAN					doStat = FALSE;
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	volatile PUINT32		Idx = FdoData->CPDRegsBase->mach_idx;

	if (FdoData->DrvState == DrvNotReady) return;

	FdoData->StatCntr += FdoData->CurStatUnitsPerCycle;

	if (FdoData->StatCntr >= UnitsPerStat) {
		 FdoData->StatCntr -= UnitsPerStat;
		 doStat = TRUE;
	}

	for (i = 0; i < 6; i++) {

		if (!mach->Axis[i].Enable) continue;

		//
		// This is to avoid second PCI access (Idx[i] is used twice.)
		//
		tmp = Idx[i];

		if (doStat) {
			mach->StatBusy = TRUE;
			mach->StatLast[i] = mach->StatHolders[i];
			mach->StatHolders[i] = 0;
			mach->StatBusy = FALSE;
		}

		mach->StatHolders[i] += ABS(mach->Axis[i].Index - tmp);

		mach->Axis[i].Index = tmp;
	}
}


// **************************************************************************
// * This makes updates to the machine spindle control. It gets called on   *
// * each timer interrupt along with updaters for I/O signals and machine   *
// * indexes.                                                               *
// *                                                                        *
// * PWM control is NOT done with actual PWM output that later has to be    *
// * converted to control voltage with an external circuitry. It has been   *
// * done this way in original Mach3 because there was no other hardware    *
// * but Parallel Ports Digital outputs there. We do have DACs on our board *
// * so we provide 0-10V signal on DAC0 output instead that can be used to  *
// * control a VFD directly without any additional circuitry. It is 10 bit  *
// * DAC so resolution is WAY better than that achievable with Mach3 PWM.   *
// **************************************************************************
VOID MachUpdateSpindle(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT		mach = FdoData->MachSharedBlock;
	volatile struct AxisInfo	*Spindle = &mach->Axis[SPINDLE];
	UINT32						tmp, newdiv, sp_pps;

	if (FdoData->DrvState == DrvNotReady) return;

	if (mach->PWM) {	// Spindle PWM control enabled
		//
		// Dac is 10 bit hence max value is 0x3ff. We do not bother
		// talking to the DAC -- it will be updated by the ISR.
		//
		//tmp = FdoData->CPDRegsBase->mach_stat;
		FdoData->DACData[0] = (0x3ff * mach->PWMFreq) / mach->PWMBaseFrequency;
	}

	if (Spindle->Enable) {	// Step/Dir Spindle
		//
		// We generate ~50% duty cycle Step Pulses. All drives make steps
		// only on low-to-high or high-to-low Step signal TRANSITION so it
		// doesn't make sense to generate fixed width pulses. That makes our
		// life easier and has an added benefit--this kind of signal can be
		// used to control some VFDs that require 50% duty cycle pulse train.
		// Pulse train is generated by counting PCI clock pulses in hardware
		// starting with newdiv value. Every time counter (counting down)
		// becomes 0 Step output is inverted. That means we need 2 such cycles
		// for a full Step signal cycle hence that ">>1" in calculation.
		//
		tmp = 0;
		if (Spindle->Jogging) {	// Do something ONLY if we are running...
			Spindle->AtSpeed = FALSE;
			if (Spindle->JoggDir != 0) tmp |= SP_ST_DIR;
			FdoData->SpindleStopping = FALSE;
			tmp |= SP_ST_RUN;
			//
			// Stopping. That is interesting case because Mach can command
			// spindle to stop and then change direction and command to run
			// BEFORE spindle had chance to stop. One can avoid an accident
			// by setting start/stop delays bigger than time required for
			// spindle to stop but it is extremely easy to misconfigure this
			// and damage an expensive spindle motor.
			//
			// That is why we LOCK into "Decelarate to Stop" mode here if
			// direction has changed while we were decelerating. That is done
			// by setting FdoData->SpindleStopping while we decelerating.
			//
			if (Spindle->Dec) {
				//
				// This will keep any Dwell from expiring until Spindle
				// is stopped. Otherwise Mach can command Spindle to start
				// in opposite direction BEFORE it stopped. No checks done
				// by Mach, it solely relies on Dwell being set long enough
				// for the Spindle to stop.
				//
				FdoData->SpindleStopping = TRUE;

				Spindle->CurVelocity -= Spindle->Acceleration;

				if (Spindle->CurVelocity <= 0) {
					FdoData->SpindleStopping = FALSE;
					Spindle->CurVelocity = 0;
					Spindle->Dec = FALSE;
					Spindle->Jogging = FALSE;
					tmp &= ~SP_ST_RUN;
				}
			//
			// Accelerate/Decelerate. Do nothing if already AtSpeed.
			//
			} else {
				if (Spindle->CurVelocity < Spindle->MaxVelocity) {
					Spindle->CurVelocity += Spindle->Acceleration;
					if (Spindle->CurVelocity > Spindle->MaxVelocity)
						Spindle->CurVelocity = Spindle->MaxVelocity;
				} else if (Spindle->CurVelocity > Spindle->MaxVelocity) {
					Spindle->CurVelocity -= Spindle->Acceleration;
					if (Spindle->CurVelocity < Spindle->MaxVelocity)
						Spindle->CurVelocity = Spindle->MaxVelocity;
				} else {
					Spindle->AtSpeed = TRUE;
				}
			}

			//
			// Don't do anything if already AtSpeed, it is all hardware.
			// Otherwise go reprogram the pulse generator.
			//
			if (!Spindle->AtSpeed) {
				if ((tmp & SP_ST_RUN) != 0) {
					sp_pps = Spindle->CurVelocity / 2048;
					sp_pps = (sp_pps > 0) ? sp_pps : 1;
					newdiv = (FdoData->PCIFreq / sp_pps - 1) >> 1;
					tmp |= SP_ST_DIV_BITS(newdiv);
				}

				FdoData->CPDRegsBase->mach_sp_ctl = tmp;
				tmp = FdoData->CPDRegsBase->mach_sp_ctl;
			}
		}
	}
}


// **************************************************************************
// * This makes updates to Encoders. MPGs have their own routine, this one  *
// * only serves 4 Encoders. This gets called on each timer interrupt along *
// * with updaters for I/O signals, machine indexes, etc.                   *
// **************************************************************************
#if 0
VOID MachUpdateEncoders(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	UINT32					tmp;
	INT						i;

	if ((tmp = FdoData->CPDRegsBase->mach_enc) != 0) {
		for (i = 0; i < 4; i++) {
			if (!mach->Encoders[i].Active) continue;
			mach->Encoders[i].Count[0] += (tmp >> (i*8)) & 0xff;
		}
	}
}
#endif


// **************************************************************************
// * This makes updates to MPGs. They are Encoders[4..6] in Mach struct. It *
// * only serves those 3 MPGs. It is pure software implementation for slow  *
// * MPG and should not be used for glass scales etc. Those should connect  *
// * to Encoders[0..3] that have hardware assistance and thus way faster.   *
// * This gets called on each timer interrupt along with updaters for I/O   *
// * signals, machine indexes, etc.                                         *
// **************************************************************************
VOID MachUpdateMPGs(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	INT						i;

	for (i = 0; i < 3; i++) {

		if (!mach->Encoders[i+4].Active) continue;

		if ((mach->Encoders[i+4].LastStateB ^ mach->Encoders[i+4].LastStateA ^
			FdoData->MPGData[i].StateA ^ FdoData->MPGData[i].StateB) != 0) {
			if ((FdoData->MPGData[i].StateA ^ mach->Encoders[i+4].LastStateB) != 0)
				mach->Encoders[i+4].Count[0]++;
			else
				mach->Encoders[i+4].Count[0]--;

		}

		mach->Encoders[i+4].LastStateA = FdoData->MPGData[i].StateA;
		mach->Encoders[i+4].LastStateB = FdoData->MPGData[i].StateB;
	}
}


// ***********************************************************************
// * This makes updates to CENTIPEDE Pulsing Engine Base Clock. This is  *
// * necessary to simulate Lathe thread cutting gearing i.e. make Z feed *
// * locked to Spindle RPM. When Spindle slows down PE Clock also slows  *
// * down the same percentage so Z Axis (actually all but Z is the only  *
// * axis that moves or matters when threading on a lathe) pulses also   *
// * come out at a slower rate. Unlike Mach3 Parallel Port driver this   *
// * one can also compensate for Spindle speedup. It works either over   *
// * the entire revolution if only index sensor is available or at every *
// * interval between spindle pulses if multi-pulse sensor is used.      *
// * Index sensor is still required even if the multi-pulse one is used  *
// * because it starts the threading pass at the exactly same position.  *
// * This gets called each Timing (or Index if there is no multi-pulse   *
// * sensor installed) interrupt when threading ONLY. It is NOT bound to *
// * timer interrupt so it won't get called if there is no Timing/Index  *
// * sensor so no interrupts from them.                                  *
// * Base PE clock is restored by the ISR when threading is finished.    *
// ***********************************************************************
VOID MachUpdatePEClock(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	INT						tmp;
	INT64 					div;

	if (FdoData->ThrSpindlePPR == 0) {
		//
		// No Reference RPM yet, do nothing. Should never happen...
		//
		return;
	}

	tmp = FdoData->SpindleCount;
	if (mach->nSlots > 1) tmp *= mach->nSlots;

	div = FdoData->BasePEClockDiv;
	div++;
	div *= tmp;
	//
	// This is for new Mach3 versions that plan threads ar 125% speed
	div *= 5;
	div /= 4;
	// End of new Mach3 version code
	//
	div /= FdoData->ThrSpindlePPR;
	div--;
	
	tmp = FdoData->CPDRegsBase->mach_ctl;
	tmp &= ~MACH_CYCLE_MASK;
	tmp |= MACH_CYCLE_BITS((INT)div);
	FdoData->CPDRegsBase->mach_ctl = tmp;
	tmp = FdoData->CPDRegsBase->mach_ctl;

	//
	// Update CurStatUnitsPerCycle so we have the right feed display
	//
//	div *= FdoData->BaseStatUnitsPerCycle;
//	div /= FdoData->BasePEClockDiv;
//	FdoData->CurStatUnitsPerCycle = (INT)div;

	//
	// Mark the base clock as adjusted, i.e. NOT at base clock rate.
	//
	FdoData->ClockAtBase = FALSE;
}

// ***********************************************************************
// ***********************************************************************
VOID MachUpdateADC(PFDO_DATA FdoData)
{
		UINT32	tmp;

		if ((FdoData->CPDRegsBase->irq_raw & CP_IRQ_ADC) != 0) {
			//
			// ADC has finished conversion, got data, and ready
			// for the next cycle.
			//
			switch (FdoData->LastADC) {
				case 0:
					FdoData->CPDRegsBase->adc = ADC_IN1;
					FdoData->ADCData[3] = FdoData->CPDRegsBase->adc;
					FdoData->LastADC = 1;
					break;
				case 1:
					FdoData->CPDRegsBase->adc = ADC_IN2;
					FdoData->ADCData[0] = FdoData->CPDRegsBase->adc;
					FdoData->LastADC = 2;
					break;
				case 2:
					FdoData->CPDRegsBase->adc = ADC_IN3;
					FdoData->ADCData[1] = FdoData->CPDRegsBase->adc;
					FdoData->LastADC = 3;
					break;
				case 3:
					FdoData->CPDRegsBase->adc = ADC_IN0;
					FdoData->ADCData[2] = FdoData->CPDRegsBase->adc;
					FdoData->LastADC = 0;
					break;
				default:
					//
					// Dummy cycle after init, data is garbage so it is
					// thrown away and Channel 0 cycle initiated.
					//
					FdoData->CPDRegsBase->adc = ADC_IN0;
					FdoData->LastADC = 0;
					break;
			}
			FdoData->CPDRegsBase->irq_raw = CP_IRQ_ADC;
			tmp = FdoData->CPDRegsBase->adc;
		}
}

// ***********************************************************************
// ***********************************************************************
VOID MachUpdateDAC(PFDO_DATA FdoData)
{
		UINT32	tmp;

		if ((FdoData->CPDRegsBase->irq_raw & CP_IRQ_DAC) != 0) {
			//
			// DAC SPI finished sending data and ready for the next cycle
			//
			switch (FdoData->LastDAC) {
				case 0:
					FdoData->CPDRegsBase->dac = DAC_B_WORD(FdoData->DACData[1]);
					FdoData->LastDAC = 1;
					break;
				case 1:
					FdoData->CPDRegsBase->dac = DAC_A_WORD(FdoData->DACData[0]);
					FdoData->LastDAC = 0;
					break;
				default:
					//
					// Should not happen but just in case -- do DAC_B cycle
					//
					FdoData->CPDRegsBase->dac = DAC_B_WORD(FdoData->DACData[1]);
					FdoData->LastDAC = 1;
					break;
			}
			FdoData->CPDRegsBase->irq_raw = CP_IRQ_DAC;
			tmp = FdoData->CPDRegsBase->irq_raw;
		}
}

// ***********************************************************************
// ***********************************************************************
VOID MachUpdateDataIn(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	static int				InData = 0;

	if (mach->DataInDone) return;
	
	if ((mach->DataInPort < ADC_0_L_DIN || mach->DataInPort > ADC_3_H_DIN) &&
		(mach->DataOutPort < PINS_0_7 || mach->DataOutPort > PINS_24_31)) {
		mach->DataInByte = 0;
	} else if (FdoData->DataInLowDone == 0) {
		switch (mach->DataInPort) {
			case ADC_0_L_DIN:
				InData = FdoData->ADCData[0];
				FdoData->DataInLowDone = ADC_0_H_DIN;
				break;
			case ADC_1_L_DIN:
				InData = FdoData->ADCData[1];
				FdoData->DataInLowDone = ADC_1_H_DIN;
				break;
			case ADC_2_L_DIN:
				InData = FdoData->ADCData[2];
				FdoData->DataInLowDone = ADC_2_H_DIN;
				break;
			case ADC_3_L_DIN:
				InData = FdoData->ADCData[3];
				FdoData->DataInLowDone = ADC_3_H_DIN;
				break;
			default:
				InData = 0;
				FdoData->DataInLowDone = 0;
				break;
		}
		
		if (FdoData->DataInLowDone == 0) {
			mach->DataInByte = 0;
		} else {
			mach->DataInByte = InData & 0xff;
		}
	} else if (FdoData->DataInLowDone == mach->DataInPort) {
		FdoData->DataInLowDone = 0;
		mach->DataInByte = (InData >> 8) & 0xff;
		InData = 0;
	} else if (mach->DataInPort == PINS_0_7) {
		mach->DataInByte = (FdoData->CPDInBuf >> 0) & 0xff;
	} else if (mach->DataInPort == PINS_8_15) {
		mach->DataInByte = (FdoData->CPDInBuf >> 8) & 0xff;
	} else if (mach->DataInPort == PINS_16_23) {
		mach->DataInByte = (FdoData->CPDInBuf >> 16) & 0xff;
	} else if (mach->DataInPort == PINS_24_31) {
		mach->DataInByte = (FdoData->CPDInBuf >> 24) & 0xff;
	} else {
		mach->DataInByte = -1;
	}

	mach->DataInDone = TRUE;
	return;
}

// ***********************************************************************
// ***********************************************************************
VOID MachUpdateDataOut(PFDO_DATA FdoData)
{
	volatile PMACH_STRUCT	mach = FdoData->MachSharedBlock;
	static int				OutData = 0;

	if (mach->DataOutDone) return;
	
	if ((mach->DataOutPort < DAC_A_L_DOUT || mach->DataOutPort > DAC_B_H_DOUT) &&
		(mach->DataOutPort < PINS_0_7 || mach->DataOutPort > PINS_24_31)) {
		goto DataOutRet;
	} else if ((mach->DataOutPort == DAC_A_L_DOUT || mach->DataOutPort == DAC_A_H_DOUT) && mach->PWM) {
		goto DataOutRet;
	} else if (FdoData->DataOutLowDone == 0) {
		switch (mach->DataOutPort) {
			case DAC_A_L_DOUT:
				OutData = mach->DataOutByte & 0xff;
				FdoData->DataOutLowDone = DAC_A_H_DOUT;
				break;
			case DAC_B_L_DOUT:
				OutData = mach->DataOutByte & 0xff;
				FdoData->DataOutLowDone = DAC_B_H_DOUT;
				break;
			default:
				break;
		}
	} else if (FdoData->DataOutLowDone == mach->DataOutPort) {
		FdoData->DataOutLowDone = 0;
		OutData |= (mach->DataOutByte << 8) & 0xff00;
		if (mach->DataOutPort == DAC_A_H_DOUT)
			FdoData->DACData[0] = OutData;
		else
			FdoData->DACData[1] = OutData;
	} else if (mach->DataOutPort == PINS_0_7) {
		FdoData->CPDOutBuf &= ~(0xff << 0);
		FdoData->CPDOutBuf |= (mach->DataOutByte & 0xff) << 0;
	} else if (mach->DataOutPort == PINS_8_15) {
		FdoData->CPDOutBuf &= ~(0xff << 8);
		FdoData->CPDOutBuf |= (mach->DataOutByte & 0xff) << 8;
	} else if (mach->DataOutPort == PINS_16_23) {
		FdoData->CPDOutBuf &= ~(0xff << 16);
		FdoData->CPDOutBuf |= (mach->DataOutByte & 0xff) << 16;
	} else if (mach->DataOutPort == PINS_24_31) {
		FdoData->CPDOutBuf &= ~(0xff << 24);
		FdoData->CPDOutBuf |= (mach->DataOutByte & 0xff) << 24;
	}

DataOutRet:
	mach->DataOutDone = TRUE;
	return;
}
